/*
 * @(#)AbacusT.c
 *
 * Copyright 2009  David A. Bagley, bagleyd@tux.org
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Used tables from How to Learn Lee's Abacus by Lee Kai-chen.
 */

#include "AbacusP.h"

#define CHAR_TO_DIGIT(c) ((c >= 'A') ? c - 'A' + 10 : c - '0')
#define DIGIT_TO_CHAR(d) ((d >= 10) ? (char) ('A' + d - 10) : (char) ('0' + d))
#define IS_DIGIT(c) ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'J'))

/* ignores decimal point and multiply signs, etc, etc */
static void simpleParser(AbacusWidget w, char * buffer,
		char* aString, char* bString, char *op)
{
	unsigned int i;
	int numberCount = 0;
	Boolean decimal = False;
	int digitCount = 0;
	int decimalPosition = w->abacus.decimalPosition;
	
	if (w->abacus.decks[BOTTOM].piece != 0)
		decimalPosition--;
	if (w->abacus.decimalPosition >= w->abacus.shiftPercent &&
			w->abacus.decks[BOTTOM].piecePercent != 0)
		decimalPosition--;
	aString[0] = '\0';
	bString[0] = '\0';
	*op = ' ';
	for (i = 0; i < strlen(buffer); i++) {
		if (buffer[i] == 'q' || buffer[i] == 'Q') {
			*op = buffer[i];
			return;
		}
		if (buffer[i] == '+' || buffer[i] == '-') {
			if (!decimal) {
				if (numberCount == 0)
					(void) sprintf(aString, "%s%c", aString, '.');
				else
					(void) sprintf(bString, "%s%c", bString, '.');
			}
			for (; digitCount < decimalPosition; digitCount++) {
				if (numberCount == 0)
					(void) sprintf(aString, "%s%c", aString, '0');
				else
					(void) sprintf(bString, "%s%c", bString, '0');
			}
			if (*op != ' ')
				return;
			*op = buffer[i];
			numberCount++;
			digitCount = 0;
			decimal = False;
		} else if (buffer[i] == '.' && !decimal) {
			decimal = True;
			if (numberCount == 0)
				(void) sprintf(aString, "%s%c", aString, '.');
			else
				(void) sprintf(bString, "%s%c", bString, '.');
		} else if (IS_DIGIT(buffer[i])) {
			if (!decimal || digitCount < decimalPosition) {
				if (numberCount == 0) {
					(void) sprintf(aString, "%s%c", aString, buffer[i]);
				} else {
					(void) sprintf(bString, "%s%c", bString, buffer[i]);
				}
			}
			if (decimal) {
				digitCount++;
			}
		}
	}
	if (!decimal) {
		if (numberCount == 0) {
			(void) sprintf(aString, "%s%c", aString, '.');
		} else {
			(void) sprintf(bString, "%s%c", bString, '.');
		}
	}
	for (; digitCount < decimalPosition; digitCount++) {
		if (numberCount == 0)
			(void) sprintf(aString, "%s%c", aString, '0');
		else
			(void) sprintf(bString, "%s%c", bString, '0');
	}
}
/* More generic */
#if 1
/* Taken from addition tables in "How to Learn Lee's Abacus */
/* Author does not really discuss carry and says to work left */ 
/* to right.  I do not know how to have that work smoothly so */ 
/* it works right to left. */
static void digitAdd(int a, int b, int base, int topFactor, int bottomNumber,
		int* lower, int* upper, int* carry)
{
	int newB = b + *carry;
	int fractBase;
	int modB, modAns, ans;
	int divA, modDivAns, divAns;

	if (bottomNumber > base / 2) /* topFactor may not be set, Russian */
		fractBase = base;
	else
		fractBase = topFactor;
	modB = newB % fractBase;
	ans = a + newB;
	modAns = ans % fractBase;
	divA = a / fractBase;
	modDivAns = (ans % base) / fractBase;
	divAns = ans / base;
	if (modAns - modB >= 0)
		*lower = modB;
	else
		*lower = modB - fractBase;
	*upper = modDivAns - divA;
	*carry = divAns;
	/*(void) printf("%d\t%d\t%d\t|\t%d\t%d\t%d\n", a, b, c, *lower, *upper, *carry);*/
}

/* Taken from subtraction tables in "How to Learn Lee's Abacus,
   made a little more generic. */
static void digitSub(int a, int b, int base, int topFactor, int bottomNumber,
		int* lower, int* upper, int* carry) {
	int newB = b - *carry;
	int fractBase;
	int modA, modB, ans;
	int divA, modDivAns, divAns;

	if (bottomNumber > base / 2) /* topFactor may not be set, Russian */
		fractBase = base;
	else
		fractBase = topFactor;
	modA = a % fractBase;
	modB = newB % fractBase;
	ans = a - newB;
	divA = a / fractBase;
	modDivAns = ((ans + base) % base) / fractBase;
	divAns = (ans - base + 1) / base;
	if (modA - modB >= 0)
		*lower = -modB;
	else
		*lower = fractBase - modB;
	*upper = modDivAns - divA;
	*carry = divAns;
	/*(void) printf("%d\t%d\t%d\t|\t%d\t%d\t%d\n", a, b, c, *lower, *upper, *carry);*/
}

#else
static void digitAdd(int a, int b, int base, int topFactor, int bottomNumber,
		int* lower, int* upper, int* carry)
{
	int newB = b + *carry;
	int fractBase;

	if (bottomNumber > base / 2) /* topFactor may not be set */
		fractBase = base;
	else
		fractBase = topFactor;
	*lower = 0;
	*upper = 0;
	*carry = 0;
	if (newB < fractBase) {
		if (a < fractBase) {
			if ((a + newB) < fractBase) { /* table 1 */
				*lower = newB;
			} else { /* table 2 */
				*lower = newB - fractBase;
				if (bottomNumber > base / 2)
					*carry = 1;
				else
					*upper = 1;
			}
		} else {
			if ((a + newB) < base) { /* table 1 */
				*lower = newB;
			} else { /* table 3 */
				*lower = newB - fractBase;
				*upper = -1;
				*carry = 1;
			}
		}
	} else {
		if ((a + newB) < base) { /* table 1 */
			*lower = newB - fractBase;
			*upper = 1;
		} else if (a >= fractBase && a + newB < base + fractBase) { /* table 4 */
			*lower = newB - fractBase;
			*upper = -1;
			*carry = 1;
		} else { /* table 3 */
			*lower = newB - base;
			*carry = 1;
		}
	}
	/*(void) printf("%d\t%d\t%d\t|\t%d\t%d\t%d\n", a, b, c, *lower, *upper, *carry);*/
}

/* Taken from subtraction tables in "How to Learn Lee's Abacus */
static void digitSub(int a, int b, int base, int topFactor, int bottomNumber,
		int* lower, int* upper, int* carry) {
	int newB = b - *carry;
	int fractBase; /* Kludge but may not be a bad fix here */
	if (bottomNumber > base / 2) /* topFactor may not be set */
		fractBase = base;
	else
		fractBase = topFactor;
	*lower = 0;
	*upper = 0;
	*carry = 0; /* if used is negative, i.e. a borrow */
	if (newB <= fractBase) {
		if (a < fractBase) {
			if ((a - newB) >= 0) { /* table 1 */
				*lower = -newB;
			} else { /* table 3 */
				*lower = fractBase - newB;
				if (bottomNumber <= base / 2)
					*upper = 1;
				*carry = -1;
			}
		} else {
			if ((a - newB) >= fractBase) { /* table 2, 3 */
				*lower = -newB;
			} else { /* table 1 */
				*lower = fractBase - newB;
				*upper = -1;
			}
		}
	} else {
		if ((a - newB) >= 0) { /* table 1 */
			*lower = fractBase - newB;
			*upper = -1;
		} else if (a < fractBase && (a - newB) >= -fractBase) { /* table 4 */
			*lower = fractBase - newB;
			*upper = 1;
			*carry = -1;
		} else { /* table 3 */
			*lower = base - newB;
			*carry = -1;
		}
	}
	/*(void) printf("%d\t%d\t%d\t|\t%d\t%d\t%d\n", a, b, c, *lower, *upper, *carry);*/
}
#endif

#ifdef DEBUG
static void testTable() {
	int i, j, k, lower, upper, carry;
#if 1
	int base = DEFAULT_BASE;
	int factor = 5;
#else
	int base = 16;
	int factor = 4;
#endif
	int c = 1;
	/*int c = 2;*/

	for (k = 0 ; k < c; k++) {
		for (j = 0 ; j < base; j++) {
			for (i = 0 ; i < base; i++) {
				carry = k;
#if 0
				digitAdd(i, j, base, factor, factor,
					&lower, &upper, &carry);
#else
				digitSub(i, j, base, factor, factor,

					&lower, &upper, &carry);
#endif
				(void) printf("%d\t%d\t|\t%d\t%d\t%d\n", i, j, lower, upper, carry);
			}
		}
	}
}
#endif

static void drawLineText(AbacusWidget w, char* line, int i)
{
#ifdef WINVER
#ifdef DEMO_FRAMED
	(void) SetTextColor(w->core.hDC, w->abacusDemo.foregroundGC);
	(void) SetBkMode(w->core.hDC, TRANSPARENT);
	(void) TextOut(w->core.hDC,
		1,
		w->core.height + w->abacusDemo.fontHeight * i - 1,
		line, (signed) strlen(line));
#else
	drawTeachText(line, i);
#endif
#else
	setAbacusText(w, ACTION_TEACH_LINE, line, i);
#endif
}

static Boolean nextPosition(AbacusWidget w, Boolean check,
		int *rPosition, int *aDigit, int *bDigit) {
	int n, rail, a, b;
	int aCount = strlen(w->abacus.aString) - 1;
	int bCount = strlen(w->abacus.bString) - 1;
	int maxCount = ((aCount >= bCount) ? aCount : bCount);
	int nonZeroOffset = 0, decimal = 0;
	int aOffset = maxCount - aCount;
	int bOffset = maxCount - bCount;

	/* skip leading zeroes but only in fractional part */
	for (n = 0; n < w->abacus.decimalPosition; n++) {
		a = CHAR_TO_DIGIT(w->abacus.aString[aCount - n]);
		b = CHAR_TO_DIGIT(w->abacus.bString[bCount - n]);
		if (a == 0 && b == 0) {
			nonZeroOffset++;
		} else {
			break;
		}
	}
	if (w->abacus.rightToLeft) {
		n = (w->abacus.step - 2) / 2;
		if (n == 0) {
			w->abacus.step += nonZeroOffset * 2;
			n = (w->abacus.step - 2) / 2;
		}
		if (n >= w->abacus.decimalPosition -
				((w->abacus.decks[BOTTOM].piece == 0) ? 0 : 1) -
				((w->abacus.decks[BOTTOM].piecePercent == 0) ? 0 : 1) -
				((checkSubdeck(w, 3)) ? 2 : 0)) {
			decimal++;
		}
		n += decimal;
		rail = n - w->abacus.decimalPosition;
		if (n > aCount)
			a = 0;
		else
			a = CHAR_TO_DIGIT(w->abacus.aString[aCount - n]);
		if (n > bCount)
			b = 0;
		else
			b = CHAR_TO_DIGIT(w->abacus.bString[bCount - n]);
		if (checkSubdeck(w, 3)) {
			rail += 2;
		} else {
			if (rail >= -1 + ((w->abacus.decks[BOTTOM].piecePercent == 0) ? 0 : 1) &&
					w->abacus.decks[BOTTOM].piece != 0)
				rail++;
			if (rail >= -w->abacus.shiftPercent - 2 &&
					w->abacus.decks[BOTTOM].piecePercent != 0)
				rail++;
			if (n >= w->abacus.decimalPosition) {
				rail--;
			}
		}
		if (!check) {
			*rPosition = rail;
			*aDigit = a;
			*bDigit = b;
		}
		return (n >= aCount && n >= bCount);
	} else {
		n = (w->abacus.step - 2) / 2;
		if (w->abacus.carryStep != 0)
			n += -(w->abacus.carryStep / 2);
		if (n >= maxCount - w->abacus.decimalPosition +
				((w->abacus.decks[BOTTOM].piece == 0) ? 0 : 1) +
				((w->abacus.decks[BOTTOM].piecePercent == 0) ? 0 : 1) +
				((checkSubdeck(w, 3)) ? 2 : 0)) {
			decimal++;
		}
		rail = maxCount - n - w->abacus.decimalPosition - 1;
		n += decimal;
		if (n - aOffset < 0)
			a = 0;
		else
			a = CHAR_TO_DIGIT(w->abacus.aString[n - aOffset]);
		if (w->abacus.carryStep == 0) {
			if (n - bOffset < 0)
				b = 0;
			else
				b = CHAR_TO_DIGIT(w->abacus.bString[n - bOffset]);
		} else {
			b = 0;
		}
		if (checkSubdeck(w, 3)) {
			rail += 3;
		} else {
			if (w->abacus.decks[BOTTOM].piece != 0 &&
					n <= maxCount - w->abacus.decimalPosition + ((w->abacus.decks[BOTTOM].piecePercent == 0) ? 0 : 1))
				rail++;
			if (n <= maxCount - w->abacus.decimalPosition + w->abacus.shiftPercent + 2 &&
					w->abacus.decks[BOTTOM].piecePercent != 0)
				rail++;
		}
		if (!check) {
			*rPosition = rail;
			*aDigit = a;
			*bDigit = b;
			if (n - aOffset >= 0 && w->abacus.step % 2 == 1)
				w->abacus.aString[n - aOffset] =
					DIGIT_TO_CHAR((a + b) % w->abacus.base);
		}
		return (maxCount - n - 1 + decimal <= nonZeroOffset);
	}
}

/* A little tricky as this is reentrant */
void teachStep(AbacusWidget w, char * buffer, int aux)
{
	char buffer1[120] = "";
	char buffer2[120] = "";
	int base = w->abacus.base;
	int topFactor = w->abacus.decks[TOP].factor;
	int bottomNumber = w->abacus.decks[BOTTOM].number;
	double a, b;

	if (w->abacus.step == 0) {
#ifdef DEBUG
		testTable();
#endif
		drawLineText(w, "Enter calculation a+b or a-b, where a positive and result positive.", 0);
		drawLineText(w, "Press enter to go through calculation steps.", 1);
#ifndef WINVER
		if (w->abacus.aString)
			free(w->abacus.aString);
		if (w->abacus.bString)
			free(w->abacus.bString);
		if ((w->abacus.aString = (char *) malloc(sizeof (char) *
				w->abacus.numDigits + 2)) &&
		   (w->abacus.bString = (char *) malloc(sizeof (char) *
				w->abacus.numDigits + 2)))
#endif
			w->abacus.step++;
	} else if (w->abacus.step == 1) {
		simpleParser(w, buffer, w->abacus.aString, w->abacus.bString,
			&(w->abacus.op));
#ifdef DEBUG
		(void) printf("buffer = %s %c %s\n",
			w->abacus.aString, w->abacus.op, w->abacus.bString);
#endif
		if (w->abacus.op == 'q' || w->abacus.op == 'Q') {
			return;
		}
		if (w->abacus.anomaly != 0) {
			addBackAnomaly(w->abacus.aString, w->abacus.anomaly,
				w->abacus.shiftAnomaly,
				base);
		}
		if (w->abacus.anomalySq != 0) {
			addBackAnomaly(w->abacus.aString, w->abacus.anomalySq,
				w->abacus.shiftAnomaly + w->abacus.shiftAnomalySq,
				base);
		}
		if (checkSubdeck(w, 3)) {
			zeroFractionalPart(w->abacus.aString);
			zeroFractionalPart(w->abacus.bString);
		}
		convertStringToAbacus(w, w->abacus.aString, aux);
		a = convertToDecimal(base, w->abacus.aString);
		b = convertToDecimal(base, w->abacus.bString);
#if 1
		if (w->abacus.op == '-' && b > a) {
			/* Revisit this, it should be allowed, but
			   rails probably need to be respected. */
			(void) sprintf(buffer1, "Subtraction overflow %s - %s",
				w->abacus.aString, w->abacus.bString);
			drawLineText(w, buffer1, 0);
			return;
		}
#endif
		if (w->abacus.op == '+' || w->abacus.op == '-') {
			(void) sprintf(buffer1, "%sing %s %c %s",
				((w->abacus.op == '+') ? "Add" : "Subtract"),
				w->abacus.aString, w->abacus.op,
				w->abacus.bString);
			drawLineText(w, buffer1, 0);
			w->abacus.step++;
		}
		w->abacus.carry = 0;
		w->abacus.carryStep = 0;
	} else if ((w->abacus.carryStep != 0 && w->abacus.carryStep % 2 == 0 &&
			w->abacus.carryStep >= 2) ||
			(w->abacus.carryStep == 0 && w->abacus.step % 2 == 0 &&
			w->abacus.step >= 2)) {
		int rPosition = 0;
		int aDigit = 0, bDigit = 0;

		(void) nextPosition(w, False, &rPosition, &aDigit, &bDigit);
		if (w->abacus.op == '+') {
			digitAdd(aDigit, bDigit, base,
				topFactor, bottomNumber,
				&(w->abacus.lower), &(w->abacus.upper),
				&(w->abacus.carry));
		} else {
			digitSub(aDigit, bDigit, base,
				topFactor, bottomNumber,
				&(w->abacus.lower), &(w->abacus.upper),
				&(w->abacus.carry));
		}
		(void) sprintf(buffer2, "For rail %d", rPosition);
		if (w->abacus.lower == 0 && w->abacus.upper == 0 &&
				w->abacus.carry == 0) {
			(void) sprintf(buffer2, "%s, do nothing", buffer2);
		} else {
			if (w->abacus.lower != 0) {
			  (void) sprintf(buffer2, "%s, %s %d",
			    buffer2,
			    ((w->abacus.lower < 0) ? "take off" : "put on"),
			    ABS(w->abacus.lower));
			  if (w->abacus.decks[BOTTOM].number <= DEFAULT_BASE / 2)
			    (void) sprintf(buffer2, "%s %s lower deck",
			      buffer2,
			      ((w->abacus.lower < 0) ? "from" : "to"));
			}
			if (w->abacus.upper != 0) {
			  (void) sprintf(buffer2,
			    "%s, %s %d %s upper deck",
			    buffer2,
			    ((w->abacus.upper < 0) ? "take off" : "put on"),
			    ABS(w->abacus.upper),
			    ((w->abacus.upper < 0) ? "from" : "to"));
			}
			if (w->abacus.carry != 0) {
			  if (w->abacus.rightToLeft) {
			    (void) sprintf(buffer2, "%s, %s %d (store in your memory)",
			      buffer2,
			      ((w->abacus.op == '+') ? "carry" : "borrow"),
			      ABS(w->abacus.carry));
			  } else {
			    (void) sprintf(buffer2, "%s, %s %d (on next move)",
			      buffer2,
			      ((w->abacus.op == '+') ? "carry" : "borrow"),
			      ABS(w->abacus.carry));
			    
			  }
			}
		}
		(void) sprintf(buffer2, "%s.", buffer2);
		drawLineText(w, buffer2, 1);
		if (w->abacus.carryStep == 0) {
			w->abacus.step++;
		} else {
			w->abacus.carryStep++;
		}
	} else {
		int rPosition = 0;
		int aDigit = 0, bDigit = 0;
		Boolean done;

		done = nextPosition(w, False, &rPosition, &aDigit, &bDigit);
		if (w->abacus.lower != 0)
			setAbacusMove(w, ACTION_MOVE, 0, 0, rPosition,
				w->abacus.lower);
		if (w->abacus.upper != 0)
			setAbacusMove(w, ACTION_MOVE, 0, 1, rPosition,
				w->abacus.upper);
			
#if 0
		if (w->abacus.carry != 0) {
			setAbacusMove(w, ACTION_MOVE, 0, 0, rPosition + 1,
				w->abacus.carry);
			w->abacus.carry = 0;
		}
#endif
		if (w->abacus.carry == 0 && !w->abacus.rightToLeft &&
				w->abacus.carryStep != 0) {
			w->abacus.carryStep = 0;
			done = nextPosition(w, True, &rPosition, &aDigit, &bDigit);
		}
		if (w->abacus.carry == 0 && w->abacus.carryStep == 0 && done) {
			drawLineText(w, "Done.", 0);
			w->abacus.step = 0;
		} else {
			if (w->abacus.rightToLeft) {
				w->abacus.step++;
			} else {
				if (w->abacus.carry != 0) {
					if (w->abacus.carryStep == 0)
						w->abacus.carryStep = 2;
					else
						w->abacus.carryStep++;
				}
				if (w->abacus.carryStep == 0)
					w->abacus.step++;
			}
		}
	}
}
